#
# This is a wrapper for the Windows SetupDI functions
# using a pure ctypes interface. This is primarily intended
# to allow finding a USB device by VID/PID and/or GUID.
#
# Copyright (c) 2008-2010 Tyler Wilson
#

import ctypes
from ctypes import wintypes
from GUID import GUID

# Windows constants
INVALID_HANDLE_VALUE = -1

# SetupDI constants
DIGCF_DEFAULT         = 0x00
DIGCF_PRESENT         = 0x02
DIGCF_ALLCLASSES      = 0x04
DIGCF_PROFILE         = 0x08
DIGCF_DEVICEINTERFACE = 0x10

# SetupDI structures
class DEVINFO_DATA(ctypes.Structure):
    _fields_ = [
        ("cbSize", wintypes.DWORD),
        ("ClassGuid", GUID),
        ("DevInst", wintypes.DWORD),
        ("Reserved", ctypes.c_void_p)
    ]

class DEVPROPKEY(ctypes.Structure):
    _fields_ = [
        ("fmtid", GUID),
        ("pid", wintypes.ULONG),
    ]
    def __init__(self, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8, pid):
        self.fmtid = GUID(fields=[l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8])
        self.pid = pid

DEVPROPTYPE = wintypes.ULONG
    
class DEVICE_INTERFACE_DATA(ctypes.Structure):
    _fields_ = [
        ("cbSize", wintypes.DWORD),
        ("InterfaceClassGuid", GUID),
        ("Flags", wintypes.DWORD),
        ("Reserved", ctypes.c_void_p)
    ]
    
class DEVICE_INTERFACE_DETAIL_DATA(ctypes.Structure):
    _fields_ = [
        ("cbSize", wintypes.DWORD),
        ("DevicePath", ctypes.c_wchar * 256)
    ]


# From devpkey.h. Tried to make it easy to move these from there to here...
DEVPKEY_NAME           = DEVPROPKEY(0xb725f130, 0x47ef, 0x101a, 0xa5, 0xf1, 0x02, 0x60, 0x8c, 0x9e, 0xeb, 0xac, 10)
DEVPKEY_Device_PDOName = DEVPROPKEY(0xa45c254e, 0xdf1c, 0x4efd, 0x80, 0x20, 0x67, 0xd1, 0x46, 0xa8, 0x50, 0xe0, 16) # DEVPROP_TYPE_STRING


# SetupDI functions (add proper args and restype too)
GetClassDevs = ctypes.windll.setupapi.SetupDiGetClassDevsW
EnumDeviceInfo = ctypes.windll.setupapi.SetupDiEnumDeviceInfo
EnumDeviceInterfaces = ctypes.windll.setupapi.SetupDiEnumDeviceInterfaces
GetDeviceInterfaceDetail = ctypes.windll.setupapi.SetupDiGetDeviceInterfaceDetailW
GetDeviceProperty = ctypes.windll.setupapi.SetupDiGetDevicePropertyW
GetCustomDeviceProperty = ctypes.windll.setupapi.SetupDiGetCustomDevicePropertyW


def FindDevicePath(guid, vendor, product):
    """This is a helper (generator) function to scan for a specific vendor and product. If found,
    a device path will be returned that can be used to open the device. If vendor and product
    are zero, search based solely on guid."""
    id = GUID(guid)
    
    #print("Looking for: vid_%04x&pid_%04x" % (vendor, product))
    
    # get the size needed to store this value
    deviceHandle = GetClassDevs(ctypes.byref(id), 0, 0, DIGCF_PRESENT|DIGCF_DEVICEINTERFACE)
#    deviceHandle = GetClassDevs(0, 0, 0, DIGCF_PRESENT|DIGCF_ALLCLASSES) #|DIGCF_DEVICEINTERFACE)
    
    if deviceHandle != -1:
        index = 0
        infoData = DEVINFO_DATA()
        infoData.cbSize = ctypes.sizeof(DEVINFO_DATA)
        
        while EnumDeviceInfo(deviceHandle, index, ctypes.byref(infoData)):
            
            # If we are using Device Classes
            
            dpt = DEVPROPTYPE()
            rSize = wintypes.DWORD()

            GetDeviceProperty(deviceHandle, ctypes.byref(infoData), ctypes.byref(DEVPKEY_Device_PDOName), ctypes.byref(dpt), 0, 0, ctypes.byref(rSize), 0)
            print("Device Property is of type %x and %d bytes in size" % (dpt.value, rSize.value))
            
            data = ctypes.create_unicode_buffer(256)
            rSize = wintypes.DWORD(ctypes.sizeof(data))

            GetDeviceProperty(deviceHandle, ctypes.byref(infoData), ctypes.byref(DEVPKEY_NAME), ctypes.byref(dpt), ctypes.byref(data), ctypes.sizeof(data), ctypes.byref(rSize), 0)
            print("Name: %s" % (data.value))
            GetDeviceProperty(deviceHandle, ctypes.byref(infoData), ctypes.byref(DEVPKEY_Device_PDOName), ctypes.byref(dpt), ctypes.byref(data), ctypes.sizeof(data), ctypes.byref(rSize), 0)
            print("PDOName: %s" % (data.value))
            
            #Try getting a custom property
            GetCustomDeviceProperty(deviceHandle, ctypes.byref(infoData), u"PortName", 0, ctypes.byref(dpt), ctypes.byref(data), ctypes.sizeof(data), 0)
            print("PortName: %s" % (data.value))

            # If we are using Device Interfaces
            
            interfaceData = DEVICE_INTERFACE_DATA()
            interfaceData.cbSize = ctypes.sizeof(DEVICE_INTERFACE_DATA)

            if EnumDeviceInterfaces(deviceHandle, ctypes.byref(infoData), ctypes.byref(id), 0, ctypes.byref(interfaceData)):
                interfaceDetail = DEVICE_INTERFACE_DETAIL_DATA()
                interfaceDetail.cbSize = ctypes.sizeof(ctypes.c_ulong) + ctypes.sizeof(ctypes.c_wchar)
                requiredSize = ctypes.c_ulong()

                GetDeviceInterfaceDetail(deviceHandle, ctypes.byref(interfaceData), 0, 0, ctypes.byref(requiredSize), 0)
                
                if GetDeviceInterfaceDetail(deviceHandle, ctypes.byref(interfaceData), ctypes.byref(interfaceDetail), requiredSize, ctypes.byref(requiredSize), 0):
                    #print("DevicePath is %s" % (interfaceDetail.DevicePath))
                    if vendor or product:
                        if interfaceDetail.DevicePath.find("vid_%04x&pid_%04x" % (vendor, product)) > -1:
                            #print("found one. moving on")
                            yield interfaceDetail.DevicePath
                    else:
                        yield interfaceDetail.DevicePath
                
            index = index + 1
            
    else:
        print("ok, something is wrong")


USB_RAW_DEVICE = "{A5DCBF10-6530-11D2-901F-00C04FB951ED}"
USB_PORTS = "{4d36e978-e325-11ce-bfc1-08002be10318}"
USB_MAESTRO = "{e0fbe39f-7670-4db6-9b1a-1dfb141014a7}"
USB_JRK = "{098da609-8fe1-410b-9fab-4d5df167d8a1}"
USB_DEVICE_TYPE = "{9d7debbc-c85d-11d1-9eb4-006008c3a19a}"
USB_POLOLU = "{6ed5eda6-46bd-40ef-85e3-8a7dae571d0e}"
USB_BUS_TYPE = "{9d7debbc-c85d-11d1-9eb4-006008c3a19a}"

if __name__ == "__main__":
    # test to find all devices of with a given device interface guid
    print("Searching for (USB_RAW_DEVICE)") 
    for dp in FindDevicePath(USB_RAW_DEVICE, 0, 0):
        print("Found a (USB_RAW_DEVICE) device at %s" % dp)        
    print("done\n")

    # Pololu Maestro Command/TTL Interface via 'USB Raw Device'
    print("Searching for (USB_RAW_DEVICE)") 
    for dp in FindDevicePath(USB_RAW_DEVICE, 0x1ffb, 0x0089):
        print("Found a (USB_RAW_DEVICE) device at %s" % dp)        
    print("done\n")
    
    print("Searching for (USB_RAW_DEVICE)") 
    for dp in FindDevicePath(USB_RAW_DEVICE, 0x06CD, 0x0121):
        print("Found a (USB_RAW_DEVICE) device at %s" % dp)        
    print("done\n")

    print("Searching for (USB_PORTS)") 
    for dp in FindDevicePath(USB_PORTS, 0, 0):
        print("Found a (USB_PORTS) device at %s" % dp)        
    print("done\n")

    print("Searching for (USB_MAESTRO)") 
    for dp in FindDevicePath(USB_MAESTRO, 0, 0):
        print("Found a (USB_MAESTRO) device at %s" % dp)        
    print("done\n")
    
    print("Searching for (USB_DEVICE_TYPE)") 
    for dp in FindDevicePath(USB_DEVICE_TYPE, 0, 0):
        print("Found a (USB_DEVICE_TYPE) device at %s" % dp)        
    print("done\n")
        
    print("Searching for (USB_POLOLU)") 
    for dp in FindDevicePath(USB_POLOLU, 0, 0):
        print("Found a (USB_POLOLU) device at %s" % dp)        
    print("done\n")
    